(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let iterative_deepening = false;;
let use_cache = true;;

type edge = int * int;;
type weight = float;;
type solution = edge list * weight;;

type cache_entry =
  | Solution of solution
  | No_solution of float
;;

module GraphHashtbl = Hashtbl.Make (struct
  type t = Graph.t
  let equal = (=)
  let compare = compare
  let hash = Graph.hash
end)

let cache = GraphHashtbl.create 1013;;

let rec fold_path_edges f path accu =
  match path with
    | u :: v :: path ->
      fold_path_edges f (v :: path) (let u, v = if u < v then u, v else v, u in f accu u v)
    | _ -> accu
;;

exception Found of int IntMap.t;;
let shortest_path g s t =
  let rec loop q pred =
    let v, q = FQueue.pop q in
    let q, pred = Graph.fold_neighbors
      (fun (q, pred) w _ ->
        if IntMap.has_key pred w
        then q, pred
        else
          let q = FQueue.push q w in
          let pred = IntMap.add pred w v in
          if w = t then raise (Found pred)
          else q, pred)
      g v (q, pred) in
    loop q pred
  in
  try
    loop (FQueue.push FQueue.empty s) (IntMap.add IntMap.empty s s)
  with Found pred ->
    let rec loop path t =
      if t = s then s :: path
      else loop (t :: path) (IntMap.get pred t)
    in
    loop [] t
;;

let fold_vertex_pairs f g accu =
  Graph.fold_vertices
    (fun accu v c_v _ ->
      Graph.fold_vertices
        (fun accu w c_w _ ->
          if v < w then f accu v c_v w c_w else accu)
        g accu)
    g accu
;;

module IntPairMap = Map.Make(struct type t = int * int let compare = compare end);;

let branching_paths g =
  let pred, q =
    Graph.fold_vertices
      (fun (pred, q) u c_u _ ->
	IntSet.fold
	  (fun (pred, q) c -> IntPairMap.add (u, c) u pred, (u, c) :: q)
	  c_u (pred, q))
      g (IntPairMap.empty, []) in
  let rec trace_path u c pred =
    let v = IntPairMap.find (u, c) pred in
    if v = u
    then [u] 
    else u :: trace_path v c pred in
  let rec loop pred q =
    Printf.eprintf "q = %a\n" (Util.output_list (fun ch (u, c) -> Printf.fprintf ch "(%d, %d)" u c)) q;
  
    if q = [] then []
    else
      let pred, q, paths =
	List.fold_left
	  (fun (pred, q, paths) (u, c) ->
	    Graph.fold_neighbors
	      (fun (pred, q, paths) v _ ->
		if IntPairMap.mem (v, c) pred then pred, q, paths
		else
		  let pred = IntPairMap.add (v, c) u pred in
		  (pred,
		   (v, c) :: q,
		   if IntSet.contains (Graph.vertex_label g v) c
		   then (
		     Printf.eprintf "found! v = %d c = %d\n%!" v c;
		     (trace_path v c pred) :: paths
		   )
		   else paths))
	      g u (pred, q, paths))
	  (pred, [], []) q
      in
      if paths <> []
      then paths
      else loop pred q
  in
  let r = loop pred q in
  Printf.eprintf "r = %a\n" (Util.output_list (Util.output_list Util.output_int)) r;
  r;
;;

let bad_paths_from_s g s max_len paths =
  (* Printf.eprintf "bad_paths_from_s %d %d\n%!" s max_len; *)
  let c_s = Graph.vertex_label g s in
  let rec trace_path t pred =
    if t = s then [s]
    else t :: trace_path (IntMap.get pred t) pred in
  let rec dfs q pred len paths =
    (* Printf.eprintf "s = %d len = %d\n%!" s len; *)
    if q = [] || len > max_len then paths
    else
      let q, pred, paths =
        List.fold_left
          (fun (q, pred, paths) v ->
            Graph.fold_neighbors
              (fun (q, pred, paths) w _ ->
                if IntMap.has_key pred w
                then (q, pred, paths)
                else
                  let pred = IntMap.add pred w v in
		  let paths = 
                    if IntSet.do_intersect (Graph.vertex_label g w) c_s
		    then (trace_path w pred) :: paths
		    else paths
                  in w :: q, pred, paths)
              g v (q, pred, paths))
          ([], pred, paths) q
      in
      dfs q pred (len + 1) paths
  in
  dfs [s] (IntMap.add IntMap.empty s s) 1 paths
;;

let is_colorful g =
  try
    ignore (Graph.fold_vertices (fun colors _ c _ -> IntSet.fold IntSet.add c colors) g IntSet.empty);
    true
  with IntSet.Already_present -> false
;;

let branching_paths g =
  let rec loop max_len =
    let paths = Graph.fold_vertices (fun paths u _ _ -> bad_paths_from_s g u max_len paths) g [] in
    (* Printf.eprintf "max_len = %d, paths = %a\n%!" max_len (Util.output_list (Util.output_list Util.output_int)) paths; *)
    if paths <> [] then paths
    else if max_len > Graph.num_vertices g
    then []
    else loop (max_len + 1)
  in
  if is_colorful g then [] else loop 1
;;

let rec list_last = function
  | [] -> failwith "list_last"
  | [x] -> x
  | _ :: xs -> list_last xs
;;

let is_branching_path g path =
  let s = List.hd path in
  let t = list_last path in
  Graph.has_vertex g s
  && Graph.has_vertex g t
  && IntSet.do_intersect (Graph.vertex_label g s) (Graph.vertex_label g t)
  && fold_path_edges (fun ok u v -> ok
    && Graph.has_vertex g u
    && Graph.has_vertex g v
    && Graph.is_connected g u v) path true
;;

exception Found_t;;
let split g s t =
  let rec bfs q seen =
    if IntSet.is_empty q
    then seen
    else
      let u, q = IntSet.pop q in
      let q, seen =
	Graph.fold_neighbors
	  (fun (q, seen) v _ ->
	    if v = t then raise Found_t;
	    if IntSet.contains seen v
	    then q, seen
	    else (IntSet.add q v, IntSet.add seen v))
	  g u (q, seen)
      in
      bfs q seen
  in
  try
    Some (bfs (IntSet.singleton s) (IntSet.singleton s))
  with Found_t -> None
;;

let rec branching_path colcom =
  match colcom.Colcom.branching_paths with
    | [] ->
      (match branching_paths colcom.Colcom.g with
	| [] -> colcom, None
	| paths -> branching_path {colcom with Colcom.branching_paths = paths})
    | path :: paths ->
      let colcom = {colcom with Colcom.branching_paths = paths} in
      if is_branching_path colcom.Colcom.g path
      then colcom, Some path
      else branching_path colcom
;;

let rec path_branches colcom =
  let colcom, path = branching_path colcom in
  match path with
    | None -> None
    | Some path ->
     Some
       (fold_path_edges
	  (fun branches u v ->
	    let dk = Graph.edge_label colcom.Colcom.g u v in
	    let g' = Graph.disconnect colcom.Colcom.g u v in
	    let colcom = {colcom with
	      Colcom.g = g';
	      Colcom.t_bridge_candidates =
		IntSet.union colcom.Colcom.t_bridge_candidates (IntSet.of_list [u; v]);} in
	    let colcoms =
	      match split g' u v with
		| None -> [colcom]
		| Some (v_u) ->
		  let g_u = Graph.subgraph g' v_u in
		  let g_v = Graph.subgraph g' (IntSet.minus (Graph.vertices g') v_u) in
		  [{colcom with Colcom.g = g_u};
	           {colcom with Colcom.g = g_v}]
	    in
	    (colcoms, dk, (fun del -> (u, v) :: del)) :: branches)
	  path [])
;;

let edge_branches colcom =
  let colcom, path = branching_path colcom in
  match path with
    | None -> None
    | Some (u :: v :: _) ->
      let colcom = {colcom with Colcom.t_bridge_candidates =
	  IntSet.union colcom.Colcom.t_bridge_candidates (IntSet.of_list [u; v]);} in
      let branch_delete =
	let dk = Graph.edge_label colcom.Colcom.g u v in
	let g' = Graph.disconnect colcom.Colcom.g u v in
	let colcom = {colcom with Colcom.g = g'} in
	let colcoms =
	  match split g' u v with
	    | None -> [colcom]
	    | Some (v_u) ->
	      let g_u = Graph.subgraph g' v_u in
	      let g_v = Graph.subgraph g' (IntSet.minus (Graph.vertices g') v_u) in
	      [{colcom with Colcom.g = g_u};
	       {colcom with Colcom.g = g_v}]
	in
	(colcoms, dk, (fun del -> (u, v) :: del)) in
      let branch_merge =
	let g', dk, restore = Colcom.merge colcom.Colcom.g v u in
	let colcom = {colcom with Colcom.g = g'} in
	([colcom], dk, restore)
      in
      Some [branch_delete; branch_merge]
    | _ -> failwith "edge_branches"
;;

let branches = ref path_branches;;
let reduce = ref Reduce.reduce_colorful_t_bridge_weighted;;

let spaces n = String.make n ' ';;

let rec branch colcom k_min k_max depth =
  let g = colcom.Colcom.g in
  if 1=0 then Printf.eprintf "%sbranch n = %d m = %d k_min = %a k_max = %a depth=%d\n%!" (spaces depth)
    (Graph.num_vertices g) (Graph.num_edges g) Util.output_float k_min Util.output_float k_max depth;
  incr Colcom.search_tree_nodes;
  let do_branching colcom k_min =
    let solution =
      let colcom, dk, restore = !reduce colcom in
      (* let colcom, dk, restore = colcom, 0.0, (fun x -> x) in *)
      let k_min = k_min -. dk in
      let k_max = k_max -. dk in
      match
	(match !branches colcom with
	  | None -> Some ([], 0.0)
	  | Some children ->
	    let rec loop best_sol = function
              | [] -> best_sol
              | (colcom', dk, restore) :: children ->
		assert (dk >= 0.0);
		let k_max' = match best_sol with
		  | None -> k_max -. dk
		  | Some (_, k) -> k -. dk -. colcom.Colcom.epsilon in
		assert (k_max' <= k_max);
		let k_min' = k_min -. dk in
		match colcom' with
		  | [colcom'] ->
		    (match branch colcom' k_min' k_max' (depth + 1) with
		       | None -> loop best_sol children
		       | Some (del, k) ->
			 let k = k +. dk in
			 assert (k >= k_min);
			 if k > k_max then loop best_sol children
			 else if k <= k_min then Some (restore del, k)
			 else
			   let best_sol = match best_sol with
			     | Some (best_del, best_k) when best_k <= k -> best_sol
			     | _ -> Some (restore del, k)
			   in loop best_sol children)
		  | [colcom1; colcom2] ->
		    let colcom1, colcom2 =
		      if Graph.num_vertices colcom1.Colcom.g < Graph.num_vertices colcom2.Colcom.g
		      then colcom1, colcom2 else colcom2, colcom1 in
		    (match branch colcom1 0.0 k_max' (depth + 1) with
		      | None -> loop best_sol children
		      | Some (del1, k1) ->
			let k_min2 = k_min' -. k1 in
			let k_max2 = k_max' -. k1 in
			(match branch colcom2 k_min2 k_max2 (depth + 1) with
			  | None -> loop best_sol children
			  | Some (del2, k2) ->
			    let k = k1 +. k2 +. dk in
			    if k > k_max then loop best_sol children
			    else if k <= k_min then Some (restore (del1 @ del2), k)
			    else 
			      let best_sol = match best_sol with
				| Some (_, best_k) when best_k <= k -> best_sol
				| _ -> Some (restore (del1 @ del2), k)
			      in loop best_sol children))
		  | _ ->
		    failwith "Branching.branch"
	    in
	    loop None children)
      with
	| None -> None
	| Some (del, k) -> Some (restore del, k +. dk)
    in
    GraphHashtbl.add cache colcom.Colcom.g
      (match solution with
	| None -> No_solution k_max
	| Some (del, k) -> Solution (del, k));
    solution
  in
  let k_min = max k_min 0.0 in
  if k_max < k_min then None
  else if not use_cache || not (GraphHashtbl.mem cache colcom.Colcom.g) then do_branching colcom k_min
  else 
    match GraphHashtbl.find cache colcom.Colcom.g with
      | Solution (del, k) when k <= k_max -> Some (del, k)
      | Solution _ -> None
      | No_solution k when k >= k_max -> None
      | No_solution k -> do_branching colcom (k +. colcom.Colcom.epsilon)
;;

let rec lower_bound colcom =
  match branching_path colcom with
    | colcom, None -> 0.0
    | colcom, Some p ->
      let w = fold_path_edges (fun w u v -> min w (Graph.edge_label colcom.Colcom.g u v)) p infinity in
      let g = fold_path_edges Graph.disconnect p colcom.Colcom.g in
      w +. lower_bound {colcom with Colcom.g = g}
;;

let solve colcom edge_branching interleaving =
  if edge_branching
  then branches := edge_branches
  else branches := path_branches;
  if not interleaving
  then reduce := fun colcom -> colcom, 0.0, fun x -> x
  else
    if edge_branching
    then reduce := Reduce.reduce_colorful_t_bridge_weighted      
    else reduce := Reduce.reduce_colorful_t_bridge;
  (* Printf.eprintf "g = %a\n" Graph.output' colcom.Colcom.g; *)
  let lb = lower_bound colcom in
  if !Util.verbose then Printf.eprintf "lower bound = %a\n%!" Util.output_float lb;
  if iterative_deepening then
    let rec loop k =
      if !Util.verbose then Printf.eprintf "k = %a\n%!" Util.output_float k;
      match branch colcom k k 0 with
	| None -> loop (k +. 1.0)
	| Some (del, cost) -> del, cost in
    loop lb
  else
    let _, ub = Heuristic.solve_merge colcom.Colcom.g in
    if !Util.verbose then Printf.eprintf "upper bound = %a\n%!" Util.output_float ub;
    match branch colcom lb ub 0 with
      | None -> failwith "Branching.solve"
      | Some (del, cost) -> del, cost
;;
