(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

type color = IntSet.t

type weight = float

type t = (color * weight IntMap.t) IntMap.t;;

let empty = IntMap.empty;;


(* Helper functions.  *)

let vertex_label g u = fst (IntMap.get g u);;
let neighbors    g u = snd (IntMap.get g u);;
let neighbor_set g u = IntMap.fold (fun s k _ -> IntSet.add s k) (neighbors g u) IntSet.empty;;
let deg          g u = IntMap.size (neighbors g u);;

let fold_vertices f g accu =
  IntMap.fold (fun accu u (l_u, n_u) -> f accu u l_u n_u) g accu
;;

let iter_vertices f g =
  IntMap.iter (fun u (l_u, n_u) -> f u l_u n_u) g
;;

let fold_edges f g accu =
  fold_vertices
    (fun accu u _ n_u ->
      IntMap.fold
        (fun accu v l_uv -> if u <= v then f accu u v l_uv else accu)
        n_u accu)
    g accu
;;
let iter_edges f g = fold_edges (fun () u v l_uv -> f u v l_uv) g ();;

let fold_vertex_pairs f g accu =
  IntMap.fold
    (fun accu u (l_u, n_u) ->
      IntMap.fold
	(fun accu v (l_v, n_v) ->
	  if u < v then f accu u l_u v l_v else accu)
	g accu)
    g accu
;;

let fold_neighbors f g u = IntMap.fold f (neighbors g u);;
let iter_neighbors f g u = IntSet.iter f (neighbors g u);;

let hash g =
  fold_vertices
    (fun hash u _ n_u -> 1143547099 * hash + 1244495467 * u + IntMap.size n_u) g (IntMap.size g)
;;


(* Querying graphs.  *)

let num_vertices = IntMap.size;;

let has_vertex g u = IntMap.has_key g u;;

let vertices g = IntMap.fold (fun vertices v _ -> IntSet.add vertices v) g IntSet.empty;;

let num_edges g =
  let num = fold_vertices (fun num u _ n_u -> num + IntMap.size n_u) g 0
  in
    assert (num mod 2 = 0);
    num / 2
;;

let num_colors g =
  let colors = fold_vertices (fun colors _ c _ -> IntSet.union colors c) g IntSet.empty in
  IntSet.size colors
;;

let is_connected g u v = IntMap.has_key (neighbors g u) v;;
let edge_label g u v = IntMap.get (neighbors g u) v;;

let is_connected_graph g =
  let rec dfs seen u =
    fold_neighbors
      (fun seen v _ ->
	if IntSet.contains seen v
	then seen
	else dfs (IntSet.add seen v) v)
      g u seen in
  if IntMap.is_empty g then true
  else
    let u = IntMap.min_key g in
    let seen = dfs (IntSet.singleton u) u in
    IntSet.size seen = num_vertices g
;;
  

(* Modifying graphs.  *)

let add_vertex g u l_u = IntMap.add g u (l_u, IntMap.empty);;
let new_vertex g c =
  let i = if IntMap.is_empty g then 0 else IntMap.max_key g + 1 in
  add_vertex g i c, i
;;

let delete_vertex g u =
  IntMap.fold
    (fun g v _ -> IntMap.modify (fun (l_v, n_v) -> (l_v, IntMap.delete n_v u)) g v)
    (neighbors g u)
    (IntMap.delete g u)
;;

let modify_vertex_label f g u =
  IntMap.modify (fun (l_u, n_u) -> (f l_u), n_u) g u
;;

let modify_edge_label f g u v =
  let g = IntMap.modify (fun (l_u, n_u) -> l_u, IntMap.modify f n_u v) g u in
  let g = IntMap.modify (fun (l_v, n_v) -> l_v, IntMap.modify f n_v u) g v in
  g
;;

let connect g u v l_uv =
  if u = v then invalid_arg "XGraph.connect: self-loop added";
  let g = IntMap.modify (fun (l_u, n_u) -> (l_u, IntMap.add n_u v l_uv)) g u in
  let g = IntMap.modify (fun (l_v, n_v) -> (l_v, IntMap.add n_v u l_uv)) g v in
  g
;;

let disconnect g u v =
  let g = IntMap.modify (fun (l_u, n_u) -> (l_u, IntMap.delete n_u v)) g u in
  let g = IntMap.modify (fun (l_v, n_v) -> (l_v, IntMap.delete n_v u)) g v in
  g
;;

let subgraph g vs =
  IntSet.fold
    (fun g' u ->
      IntMap.add g' u (vertex_label g u, IntMap.filter (fun v _ -> IntSet.contains vs v) (neighbors g u)))
    vs empty
;;

let fold_connected_components f g accu =
  let accu, _ =
    fold_vertices
      (fun (accu, seen) u _ n_u ->
	if IntSet.contains seen u
	then accu, seen
	else
	  let rec dfs seen component u =
	    fold_neighbors
	      (fun (seen, component) v _ ->
		if IntSet.contains seen v
		then seen, component
		else
		  let seen = IntSet.add seen v in
		  let component = IntSet.add component v in
		  dfs seen component v)
	      g u (seen, component) in
	  let seen, component = dfs (IntSet.add seen u) (IntSet.singleton u) u in
	  f accu (subgraph g component), seen)
      g (accu, IntSet.empty)
  in
  accu
;;

let merge g u v =
  let g' = delete_vertex g v in
  let g' = fold_neighbors
    (fun g' w w_vw ->
      if w = u then g'
      else if is_connected g u w
      then modify_edge_label (fun w_uw -> w_uw +. w_vw) g' u w
      else connect g' u w w_vw)
    g v g' in
  g'
;;

let connectivity g =
  let most_tight g u =
    let w, _ = 
      fold_neighbors
	(fun (w, w_uw) v w_uv -> if w_uv > w_uw then v, w_uv else w, w_uw)
	g u (-1, neg_infinity) in
    w in
  let rec minimum_cut_phase g u =
    let rec loop g u =
      if num_vertices g = 2
      then
	let s = u in
	let t = IntMap.min_key (IntMap.delete g s) in
	s, t, edge_label g t s
      else
	let v = most_tight g u in
	let g = merge g v u in
	minimum_cut_phase g v in
    loop g u in
  let rec minimum_cut g u min_cut =
    if num_vertices g <= 1 then min_cut
    else
      let s, t, cut = minimum_cut_phase g u in
      let min_cut = min min_cut cut in
      let g = merge g t s in
      minimum_cut g t min_cut in
  minimum_cut g (IntMap.min_key g) infinity
;;

let cut g vs =
  IntSet.fold
    (fun cut u ->
      IntMap.fold
	(fun cut v w_uv -> if not (IntSet.contains vs v) then cut +. w_uv else cut)
	(neighbors g u) cut)
    vs 0.0
;;


(* I/O  *)

let output' output_v output_lv output_le channel g =
  Printf.fprintf channel "# n = %d, m = %d\n" (num_vertices g) (num_edges g);
  iter_vertices (fun u l_u _ -> Printf.fprintf channel "v %a %a\n" output_v u output_lv l_u) g;
  iter_edges (fun u v l_uv   -> Printf.fprintf channel "e %a %a %a\n" output_v u output_v v output_le l_uv) g;
;;

let output channel g =
  output'
    Util.output_int
    (fun ch colors -> IntSet.iter (fun c -> Printf.fprintf ch "%d " c) colors)
    Util.output_float
    channel
    g
;;

let max_colors g =
  fold_vertices (fun m _ c_v _ -> max m (IntSet.size c_v)) g 0
;;

let palette = [|"#e50000"; "#034cff"; "#15b01a"; "#ffff14"; "#f97306"; "#7e1e9c";
		"#ff81c0"; "#029386"; "#c20078"; "#653700"; "#95d0fc"; "#06470c";
		"#929591"; "#33ffff"; "#ceb301"; "#d8dcd6"; "#cea2fd"|];;

let output_dot channel g =
  let h = 0.23 +. 0.19 *. ceil (sqrt (float_of_int (max_colors g))) in
  Printf.fprintf channel "graph {\nnode [shape=circle,height=%f,fixedsize=yes];\n" h;
  iter_vertices
    (fun u c_u _ ->
      Printf.fprintf channel "%d [label=<<table border=\"0\" cellborder=\"0\" cellspacing=\"0\" cellpadding=\"0\"><tr><td>%d</td></tr><tr>" u u;
      let l = int_of_float (ceil (sqrt (float_of_int (IntSet.size c_u)))) in
      ignore
	(IntSet.fold
	   (fun i c ->
	     let i = if i > l then (Printf.fprintf channel "</tr><tr>"; 1) else i + 1 in
	     output_string channel ("<td><font color=\"" ^ palette.(c) ^ "\">●</font></td>"); i)
	   c_u 0);
      output_string channel "</tr></table>>];\n")
    g;
  iter_edges
    (fun u v w_uv -> Printf.fprintf channel "%d -- %d [label=%s];\n" u v (Util.string_of_float w_uv))
    g;
  output_string channel "}\n";
;;

let output_pdf channel g =
  let dot_file, dot_out = Filename.open_temp_file "colcom" ".dot" in
  let pdf_file = Filename.temp_file "colcom" ".pdf" in
  output_dot dot_out g;
  close_out dot_out;
  (match Unix.system ("/usr/bin/neato -Gsep=0.5 -Gsplines=true -Goverlap=scale -Tpdf " ^ dot_file ^ " -o" ^ pdf_file ^ " 2> /dev/null") with
    | Unix.WEXITED 0 -> ()
    | _ -> Printf.eprintf "neato failed\n%!"; exit 1);
  Unix.unlink dot_file;
  Util.output_file channel pdf_file;
  Unix.unlink pdf_file;
;;

let show g =
  let pdf_file, pdf_out = Filename.open_temp_file "colcom" ".pdf" in
  output_pdf pdf_out g;
  close_out pdf_out;
  (match Unix.system ("evince " ^ pdf_file ^ " &") with
    | Unix.WEXITED 0 -> ()
    | _ -> Printf.eprintf "evince failed\n%!"; exit 1);
;;
