(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** Undirected functional graphs with colored vertices and weighted
    edges. Vertices and colors are represented by nonnegative
    integers.  No multiple edges or self-loops are possible.  *)

(** The type of a vertex color.  *)
type color = IntSet.t

(** The type of an edge weight.  *)
type weight = float

(** The type of a graph.  *)
type t

(** The empty graph.  *)
val empty : t


(** Querying graphs.  *)

(** Number of vertices. O(1) time.  *)
val num_vertices : t -> int

(** Number of edges. O(n) time.  *)
val num_edges : t -> int

(** Number of colors. O(n) time.  *)
val num_colors : t -> int

(** [has_vertex g i] returns true if there is a vertex [i] in [g].
    O(log n) time.  *)
val has_vertex : t -> int -> bool

(** Returs the set of vertices. O(n log n) time.  *)
val vertices : t -> IntSet.t

(** [deg g i] returns the number of neighbors of [i] in [g].
    O(log n) time.  *)
val deg : t -> int -> int

(** Color of a vertex. O(log n) time.  *)
val vertex_label : t -> int -> color

(** Weight of an edge. O(log n) time.  *)
val edge_label : t -> int -> int -> weight

(** Returns true if two vertices are connected.  O(log n) time.  *)
val is_connected : t -> int -> int -> bool

(** [neighbors g u] returns a mapping from neighbors of [u] in [g] to
    the corresponding edge weight.  O(log n) time.  *)
val neighbors : t -> int -> weight IntMap.t

(** The set of neighbors. O(n) time.  *)
val neighbor_set : t -> int -> IntSet.t

(** True if there is at most one connected component.  *)
val is_connected_graph : t -> bool

(** [fold_vertices f g a] computes [(f iN cN nN ... (f i1 c1 n1
    a)...)], where [i1, ... iN] are the vertices of [g], [cN] is the
    color of [iN], and [nN] is the map of neighbors of [iN].
    O(n) time.  *)
val fold_vertices :
  ('a -> int -> color -> weight IntMap.t -> 'a) -> t -> 'a -> 'a

(** [iter_vertices f g] calls [f u cu nu] for each vertex [u] in [g],
    where [cu] is the color of u and [nu] is the set of neighbors
    of [u].  *)
val iter_vertices :
  (int -> color -> weight IntMap.t -> unit) -> t -> unit

(** [fold_edges f g a] computes [(f iN jN wN ... (f i1 j1 w1 a)...)],
    where [(i1, j1) ... (iN, jN)] are the edges of [g], with weight [wN].
    O(m) time.  *)
val fold_edges : ('a -> int -> int -> weight -> 'a) -> t -> 'a -> 'a

(** [iter_edges f g a] calls [f i1 j1 w1, ..., f iN jN wN] where
    [(i1, j1) ... (iN, jN)] are the edges of [g], with weight [wN].
    O(m) time.  *)
val iter_edges : (int -> int -> weight -> unit) -> t -> unit

(** [fold_vertices f g a] computes [(f iN cN jN dN ... (f i1 c1 j1 c1 a)...)],
    where [i1, ... iN] and [j1, ... jN] are the vertices of [g] with
    [i] < [j], [cN] is the color of [iN], and [dN] is the color of [jN].
    O(n^2) time.  *)
val fold_vertex_pairs :
  ('a -> int -> color -> int -> color -> 'a) -> t -> 'a -> 'a

(** [fold_neighbors f g i a] computes [(f jN wN ... (f j1 w1 a)...)],
    where [j1, ..., jN] are the neighbors of [i] in [g], with weights
    [w1, ... wn].  O(log n + deg i) time.  *)
val fold_neighbors :
  ('a -> int -> weight -> 'a) -> t -> int -> 'a -> 'a

(** [fold_connected_components f g a] computes [(f cN ... (f c1 a)...)],
    where [c1, ..., cN] are the connected components of [g].
    O(m \log m) time.  *)
val fold_connected_components :
  ('a -> t -> 'a) -> t -> 'a -> 'a

(** Subgraph induced by a vertex set.  *)
val subgraph : t -> IntSet.t -> t

(** Minimum weight of edges that need to be deleted to cut the
    connected graph into two components.  *)
val connectivity : t -> float

(** [cut g vs] yields the total weight of edges in [g] with exactly
    one endpoint in [vs].  *)
val cut : t -> IntSet.t -> float

(** A simple hash based only on vertex degrees.  *)
val hash : t -> int


(** Modifying graphs.  *)

(** [add_vertex g i c] returns [g] with an additional new vertex [i]
    of color [c]. Raises [IntMap.Already_present] if [i] is already a
    vertex.  O(log n) time.  *)
val add_vertex : t -> int -> color -> t

(** [new_vertex g c] returns [g] with an additional new vertex [i] of
    color [c], and the new vertex [i].  O(log n) time.  *)
val new_vertex : t -> color -> t * int

(** [delete_vertex g i] returns [g] with vertex [i] deleted.
    O(deg(i) log(n)) time.  *)
val delete_vertex : t -> int -> t

(** [connect g u v w] returns [g] with vertices [u] and [v] connected
    with edge weight [w].  Raises [Not_found] when [u] or [v] do not
    exist in [g]. Raises [IntSet.Already_present] when [u] and [v] are
    already connected.  Raises [Invalid_argument] when [u = v].
    O(log n) time.  *)
val connect : t -> int -> int -> weight -> t

(** [disconnect g u v] returns [g] with vertices [u] and [v]
    disconnected. Raises [Not_found] when [u] or [v] do not exist in
    [g], or when [u] and [v] are not connected.  O(log n) time.  *)
val disconnect : t -> int -> int -> t

(** [modify_vertex_label f g u] changes the color [c] of [u] in [g] to
    [f c].  O(log n) time.  *)
val modify_vertex_label : (color -> color) -> t -> int -> t

(** [modify_edge_label f g u v] changes the weight [w] of [{u, v}] in
    [g] to [f w].  O(log n) time.  *)
val modify_edge_label : (weight -> weight) -> t -> int -> int -> t


(** I/O  *)

(** [output c g] prints a debug representation of [g] to channel [c].
    O(m) time.  *)
val output : out_channel -> t -> unit

(** [output c g] writes a drawing of [g] in PDF format to channel [c].
    Requires GraphViz.  O(m) time.  *)
val output_pdf : out_channel -> t -> unit
