(* colcom -- solve the colorful components problem
   Copyright (C) 2011  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** Sets of nonnegative integers.  *)
    
(** The type of the set.  *)
type t

(** Exception  *)
exception Already_present

(** The empty set.  *)
val empty : t

(** Returns a one-element set.  *)
val singleton : int -> t

(** Returns a set from the elements of a list. Raises
    [Already_present] on duplicate entries.  *)
val of_list : int list -> t


(** Querying sets.  *)

(** Test whether the set is empty. O(1) time.  *)
val is_empty : t -> bool

(** Returns the number of elements. O(1) time.  *)
val size : t -> int

(** Returns the minimal element. O(log n) time.  *)
val min : t -> int

(** Returns the maximal element. O(log n) time.  *)
val max : t -> int
  
(** [choose s] returns an arbitrary element of [s], or raises
    [Not_found] when [s] is empty. O(log n) time.  *)
val choose : t -> int

(** [contains s i] returns true if [i] is element of [s].  Raises
    [Invalid_argument] when [i] is negative. O(log n) time.  *)
val contains : t -> int -> bool

(** [is_subset s1 s2] tests whether the set [s1] is a subset of the
    set [s2].  *)
val is_subset : t -> t -> bool

(** [fold f s a] computes [(f iN ... (f i1 a)...)], where [i1 ... iN]
    are the elements of [s]. O(n t) time, where [t] is the time bound
    for [f].  *)
val fold : ('a -> int -> 'a) -> t -> 'a -> 'a

(** [iter f s] calls [f i1, ..., f iN], where [i1 ... iN] are the
    elements of [s].  *)
val iter : (int -> unit) -> t -> unit

(** [for_all p s] returns true if [p x] holds for all [x] in [s].  *)
val for_all : (int -> bool) -> t -> bool

(** [find_opt p s] returns [Some i] when [p] returns [Some i] for some
    element of [s], or [None] if [p] returns [None] for all elements of
    [s].  *)
val find : (int -> 'a option) -> t -> 'a option

(** [filter p s] retains a set containing those elements [x] from [m]
    for which [p x] returns true.  *)
val filter : (int -> bool) -> t -> t


(** Modifying sets.  *)

(** [add s i] returns a set containing all elements of [s], plus [i].
    Raises [Already_present] if [i] was already in [s]. Raises
    [Invalid_argument] when [i] is negative. O(log n) time.  *)
val add : t -> int -> t

(** [put s i] returns a set containing all elements of [s], plus
    [i]. If [i] was already in [s], [s] is returned unchanged.  Raises
    [Invalid_argument] when [i] is negative. O(log n) time.  *)
val put : t -> int -> t

(** [delete s i] returns a set containing all elements of [s], minus
    [i]. Raises [Not_found] if [i] is not in [s]. Raises
    [Invalid_argument] when [i] is negative. O(log n) time.  *)
val delete : t -> int -> t

(** [remove s i] returns a set containing all elements of [s], minus [i].
    Raises [Invalid_argument] when [i] is negative. O(log n) time.  *)
val remove : t -> int -> t

(** [pop s] returns an arbitrary element of [s] and the set [s]
    without this element. Raises [Not_found] when [s] is empty.
    O(log n) time.  *)
val pop : t -> int * t


(** Combining sets.  *)

(** Set union.  *)
val union : t -> t -> t

(** Set intersection.  *)
val intersection : t -> t -> t

(** True if the two sets have a nonempty intersection.  *)
val do_intersect : t -> t -> bool

(** Number of elements in the intersection.  *)
val intersection_size : t -> t -> int

(** Set difference.  *)
val minus : t -> t -> t


(** I/O  *)

(** [output c s] prints a debug representation of [s] to channel [c].
    O(n) time.  *)
val output : out_channel -> t -> unit
