(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let usage_msg = "Solve the colorful components problem";;

type solver =
  | Path_branching
  | Edge_branching
  | Merge_heuristic
;;

let solver = ref Path_branching;;

let interleaving      = ref true;;
let data_reduction    = ref true;;

let stats_only	      = ref false;;
let random_params     = ref false;;
let dump_components   = ref false;;
let del_out           = ref stdout;;
let components_out    = ref None;;

let specs = [
  ("-e",   Arg.Unit(fun () -> solver := Edge_branching),
	   "Use edge branching solver");
  ("-h",   Arg.Unit(fun () -> solver := Merge_heuristic),
	   "Use merging heuristic");
  ("-n",   Arg.Clear(data_reduction),
	   "No initial data reduction");
  ("-l",   Arg.Clear(interleaving),
	   "No interleaved data reduction");
  ("-s",   Arg.Set(stats_only),
	   "Print statistics only");
  ("-r",   Arg.Set(random_params),
	   "Show random model parameters per connected component");
  ("-o",   Arg.String(fun file -> del_out := open_out file),
	   "Print deleted edges to file (default: stdout)");
  ("-c",   Arg.String(fun file -> components_out := Some (open_out file)),
	   "Print remaining colorful components to file");
  ("-v",   Arg.Set(Util.verbose),
	   "Verbose output");
  ("-d",   Arg.Set(dump_components),
	   "Dump irreducible components as .pdf");
];;


external read_and_fold_components :
  ('a -> (int * string) array -> (int * int * float) list -> 'a)
  -> 'a -> bool -> 'a = "read_and_fold_components";;

let decode_cc_graph vertices edges =
  let rec ctz x = if x land 1 <> 0 then 0 else 1 + (ctz (x lsr 1)) in
  let colorset code =
    let rec loop colors code =
      if code = 0 then colors
      else loop (IntSet.add colors (ctz code)) (code land (code - 1)) in
    loop IntSet.empty code in
  let vertices = Array.mapi (fun i (code, name) -> (i, colorset code, name)) vertices in
  let g = Array.fold_left (fun g (i, c, name) -> Graph.add_vertex g i c) Graph.empty vertices in
  let g = List.fold_left (fun g (u, v, w) -> Graph.connect g u v w) g edges in
  let vertex_names = Array.fold_left
    (fun vertex_names (i, _, name) ->
      IntMap.add vertex_names i name) IntMap.empty vertices in
  (* Printf.eprintf "g = %a\n%!" Graph.output' g; *)
  g, vertex_names
;;

let total_n = ref 0;;
let total_m = ref 0;;
let total_k = ref 0.0;;
let total_c = ref 0;;

let red_n = ref 0;;
let red_m = ref 0;;
let red_c = ref 0;;

let max_n = ref 0;;
let max_m = ref 0;;
let max_c = ref 0;;
let max_k = ref 0.0;;

let max_red_n = ref 0;;
let max_red_m = ref 0;;
let max_red_c = ref 0;;

let dump_nr = ref 1;;

let is_colorful_component g =
  try
    ignore (Graph.fold_vertices
	      (fun colors _ c _ ->
		if IntSet.do_intersect colors c
		then raise IntSet.Already_present
		else IntSet.union colors c)
	      g IntSet.empty);
    true
  with IntSet.Already_present -> false
;;

let is_colorful g =
  Graph.fold_connected_components (fun ok comp -> ok && is_colorful_component comp) g true
;;

let is_solution g del =
  is_colorful (List.fold_left (fun g' (u, v) -> Graph.disconnect g' u v) g del)
;;

let (@@) f g = fun x -> f (g x);;

let handle_component () vertices edges =
  let g, vertex_names = decode_cc_graph vertices edges in
  if !Util.verbose then Printf.eprintf "connected component n = %d m = %d c = %d\n%!"
    (Graph.num_vertices g) (Graph.num_edges g) (Graph.num_colors g);
  let start = Util.timer () in
  let colcom = Colcom.make g 1.0 in
  let colcom, dk, restore =
    if not !data_reduction
    then colcom, 0.0, fun x -> x
    else
      let colcom, dk, restore = Reduce.forbidden_edge colcom in  
      let colcom, dk', restore' =
	if !solver = Path_branching
	then Reduce.reduce_colorful_t_bridge          colcom
	else Reduce.reduce_colorful_t_bridge_weighted colcom in
      colcom, dk +. dk', restore @@ restore' in
  if !Util.verbose && !data_reduction then Printf.eprintf "reduced             n = %d m = %d c = %d\n%!"
    (Graph.num_vertices colcom.Colcom.g) (Graph.num_edges colcom.Colcom.g) (Graph.num_colors colcom.Colcom.g);
  if !dump_components && Graph.num_vertices colcom.Colcom.g > 0 then begin
    let filename = Printf.sprintf "irred%04d.pdf" !dump_nr in
    let f = open_out filename in
    Graph.output_pdf f colcom.Colcom.g;
    close_out f;
    if !Util.verbose then Printf.eprintf "wrote %s\n%!" filename;
    incr dump_nr;
  end;
  let del, weight =
    if Graph.num_vertices colcom.Colcom.g <= 1 then [], 0.0
    else match !solver with
      | Edge_branching -> Branching.solve colcom true  !interleaving
      | Path_branching -> Branching.solve colcom false !interleaving
      | Merge_heuristic -> Heuristic.solve_merge colcom.Colcom.g in
  let del = restore del in
  let weight = weight +. dk in
  let stop = Util.timer () in
  if !Util.verbose && Graph.num_vertices colcom.Colcom.g > 1 then Printf.eprintf "time %f\n%!" (stop -. start);
  if !stats_only then begin
    total_n := !total_n + Graph.num_vertices g;
    total_m := !total_m + Graph.num_edges g;
    total_c := max !total_c
      (Graph.fold_vertices
	 (fun max_c _ c _ -> max max_c (if IntSet.is_empty c then 0 else IntSet.max c + 1)) g 0);
    total_k := !total_k +. weight; 
    red_n := !red_n + Graph.num_vertices colcom.Colcom.g;
    red_m := !red_m + Graph.num_edges colcom.Colcom.g;
    red_c := max !red_c (Graph.num_colors colcom.Colcom.g);
    if Graph.num_edges g > !max_m then begin
      max_n := Graph.num_vertices g;
      max_m := Graph.num_edges g;
      max_c := Graph.num_colors g;
      max_k := weight;
      max_red_n := Graph.num_vertices colcom.Colcom.g;
      max_red_m := Graph.num_edges colcom.Colcom.g;
      max_red_c := Graph.num_colors colcom.Colcom.g;
    end;
  end else
    List.iter
      (fun (u, v) -> Printf.fprintf !del_out "%s %s\n" (IntMap.get vertex_names u) (IntMap.get vertex_names v))
      del;
  (match !components_out with | None -> ()
    | Some out ->
      let g' = List.fold_left (fun g' (u, v) -> Graph.disconnect g' u v) g del in
      Graph.fold_connected_components
	(fun () g_c ->
	  ignore (Graph.fold_vertices
		    (fun first u _ _ ->
		      Printf.fprintf out "%s%s" (if first then "" else " ") (IntMap.get vertex_names u); false)
		    g_c true);
	  Printf.fprintf out "\n")
	g' ());
  if !random_params then begin
    let choose_2 x = (x * (x - 1)) / 2 in
    let (//.) x y = (float_of_int x) /. (float_of_int y) in
    let g' = List.fold_left (fun g' (u, v) -> Graph.disconnect g' u v) g del in
    let ccs = Graph.fold_connected_components (fun ccs cc -> cc :: ccs) g' [] in
    let n = Graph.num_vertices g' in
    let n_c = Graph.num_colors g' in
    let p = List.length ccs in
    let p_v = n //. (n_c * p) in
    let m_max = choose_2 n in
    let m_internal_max = List.fold_left (fun m_internal_max g_c -> m_internal_max + choose_2 (Graph.num_vertices g_c)) 0 ccs in
    let m_external_max = m_max - m_internal_max in
    let m_internal = Graph.num_edges g' in
    let m_external = List.length del in
    let p_e = m_internal //. m_internal_max in
    let p_x = m_external //. m_external_max in
    if p > 1 then
      Printf.printf "%d %d %f %f %f\n%!" n_c n p_v p_e p_x;
  end;
  if not (is_solution g del)
  then begin
    flush_all ();
    Printf.eprintf "Internal error: solution verification failed\n%!";
    exit 1;
  end;
  let actual_weight = List.fold_left
    (fun actual_weight (u, v) -> actual_weight +. Graph.edge_label g u v) 0.0 del in
  if abs_float (weight -. actual_weight) > 1e-6 then begin
    Printf.eprintf "%!Internal error: weight returned %f, actually %f\n%!" weight actual_weight;
    exit 1;
  end;
;;

let main () =
  Arg.parse specs (fun _ -> Arg.usage specs usage_msg) usage_msg;
  let start = Util.timer () in
  read_and_fold_components handle_component () !Util.verbose;
  let stop = Util.timer () in
  if !stats_only then
    Printf.printf
      "%5d %5d %3d %8s %5d %5d %3d %5d %5d %3d %8s %5d %5d %3d %7.2f %8d\n%!"
      !total_n   !total_m   !max_c     (Util.string_of_float !total_k)
      !red_n     !red_m     !red_c
      !max_n     !max_m     !max_c     (Util.string_of_float !max_k)
      !max_red_n !max_red_m !max_red_c
      (stop -. start) !Colcom.search_tree_nodes
;;

main ();;

