(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let (@@) f g = fun x -> f (g x);;

let fold_colorful_connected_subgraphs f g accu =
  let n_excl v vs =
    let n_v = Graph.neighbor_set g v in
    let n_vs = IntSet.fold (fun n_vs u -> IntSet.union n_vs (Graph.neighbor_set g u)) vs vs in
    IntSet.minus n_v n_vs in
  let rec extend_subgraph v_subgraph c_subgraph v_extension v accu =
    let accu = f accu v_subgraph in
    let rec loop accu v_extension =
      if IntSet.is_empty v_extension then accu
      else
	let w, v_extension = IntSet.pop v_extension in
	let c_w = Graph.vertex_label g w in
	if IntSet.do_intersect c_subgraph c_w
	then loop accu v_extension
	else
          let v_extension' = IntSet.union v_extension (IntSet.filter (fun u -> u > v) (n_excl w v_subgraph)) in
          let accu =
            extend_subgraph
              (IntSet.add v_subgraph w) (IntSet.union c_subgraph c_w) v_extension' v accu in
          loop accu v_extension in
    loop accu v_extension in
  Graph.fold_vertices
    (fun accu v c_v n_v ->
      let v_extension =
	IntMap.fold
          (fun v_extension u _ -> if u > v then IntSet.add v_extension u else v_extension)
          n_v IntSet.empty
      in
      extend_subgraph (IntSet.singleton v) c_v v_extension v accu)
    g accu
;;

exception Found of IntSet.t;;
let dense_colorful_component g =
  try
    fold_colorful_connected_subgraphs
      (fun () vs ->
       if IntSet.size vs > 1
       then
         let cut = Graph.cut g vs in
         let connectivity = Graph.connectivity (Graph.subgraph g vs) in
         if connectivity >= cut
         then raise (Found (vs)))
      g ();
    None
  with
    | Found vs ->
      let rec loop g dk restore vs =
       if IntSet.size vs <= 1
       then Some (g, dk, restore)
       else
         let u, vs = IntSet.pop vs in
         let v = IntSet.min vs in
         let g', dk', restore' = Colcom.merge g v u in
         loop g' (dk +. dk') (restore @@ restore') vs
      in
      loop g 0.0 (fun x -> x) vs
;;

let reduce_colorful_t_bridge colcom =
  let g = colcom.Colcom.g in
  let candidates = colcom.Colcom.t_bridge_candidates in
  let weight_in_out g s u =
    Graph.fold_neighbors
      (fun (w_in, w_out) v w_uv ->
	if IntSet.contains s v
	then w_in +. w_uv,  w_out
	else w_in, w_out +. w_uv)
      g u (0.0, 0.0) in
  let rec loop g dk del candidates =
    if IntSet.is_empty candidates
    then g, dk, del
    else
      let u, candidates = IntSet.pop candidates in
      let rec search s c_s n =
	let c_n = IntSet.fold (fun c_n v -> IntSet.union c_n (Graph.vertex_label g v)) n IntSet.empty in
	let cut = IntSet.fold
	  (fun cut u ->
	    (Graph.fold_neighbors (fun cut v w_uv -> if IntSet.contains s v then cut else cut +. w_uv) g u cut))
	  s 0.0 in
	(* Printf.eprintf "u = %d s = %a c_s = %a n = %a c_n = %a cut = %f\n%!" *)
	(*   u IntSet.output s IntSet.output c_s IntSet.output n IntSet.output c_n cut; *)
	if IntSet.is_subset c_n c_s && Graph.connectivity (Graph.subgraph g s) >= cut
	then Some (s, n, cut)
	else
	  let best_new, _, _ =
	    IntSet.fold
	      (fun (best_new, best_in, best_out) v ->
		let v_in, v_out = weight_in_out g s v in
		if v_out -. v_in /. 10.0 < best_out -. best_in /. 10.0
		then v, v_in, v_out
		else best_new, best_in, best_out)
	      (IntSet.filter (fun v -> not (IntSet.do_intersect c_s (Graph.vertex_label g v))) n)
	      (-1, 0.0, infinity) in
	  if best_new = -1 then None
	  else
	    let s = IntSet.add s best_new in
	    let c_s = IntSet.union c_s (Graph.vertex_label g best_new) in
	    let n = IntSet.delete n best_new in
	    let n = IntSet.union n (IntSet.minus (Graph.neighbor_set g best_new) s) in
	    search s c_s n
      in
      if not (Graph.has_vertex g u) then loop g dk del candidates else
      match search (IntSet.singleton u) (Graph.vertex_label g u) (Graph.neighbor_set g u) with
	| None -> loop g dk del candidates
	| Some (s, n, cut) ->
	  (* Printf.eprintf "found colorful %a-bridge!\n%!" Util.output_float cut; *)
	  let cut_edges = IntSet.fold
	    (fun cut_edges u ->
	      (Graph.fold_neighbors (fun cut_edges v _ ->
		if IntSet.contains s v then cut_edges else (u, v) :: cut_edges) g u cut_edges))
	    s [] in
	  let g = IntSet.fold (fun g u -> Graph.delete_vertex g u) s g in
	  let candidates = IntSet.union candidates n in
	  loop g (dk +. cut) (cut_edges @ del) candidates
  in
  let g, dk, del = loop g 0.0 [] candidates in
  {colcom with Colcom.g = g; Colcom.t_bridge_candidates = IntSet.empty}, dk, (fun del' -> del @ del')
;;

(* TODO: unify with flag for weighted/unweighted *)
let reduce_colorful_t_bridge_weighted colcom =
  let g = colcom.Colcom.g in
  let candidates = colcom.Colcom.t_bridge_candidates in
  (* Printf.eprintf "candidates = %a g = %a\n" IntSet.output candidates Graph.output' g; *)
  let weight_in_out g s u =
    Graph.fold_neighbors
      (fun (w_in, w_out) v w_uv ->
	if IntSet.contains s v
	then w_in +. w_uv,  w_out
	else w_in, w_out +. w_uv)
      g u (0.0, 0.0) in
  let rec loop g dk restore candidates =
    if IntSet.is_empty candidates
    then g, dk, restore
    else
      let u, candidates = IntSet.pop candidates in
      let rec search s c_s n =
	let cut = IntSet.fold
	  (fun cut u ->
	    (Graph.fold_neighbors (fun cut v w_uv -> if IntSet.contains s v then cut else cut +. w_uv) g u cut))
	  s 0.0 in
	(* Printf.eprintf "u = %d s = %a c_s = %a n = %a cut = %f\n%!" *)
	(*   u IntSet.output s IntSet.output c_s IntSet.output n cut; *)
	if IntSet.size s > 1 && Graph.connectivity (Graph.subgraph g s) >= cut
	then Some (s, n, cut)
	else
	  let best_new, _, _ =
	    IntSet.fold
	      (fun (best_new, best_in, best_out) v ->
		let v_in, v_out = weight_in_out g s v in
		if v_out -. v_in /. 10.0 < best_out -. best_in /. 10.0
		then v, v_in, v_out
		else best_new, best_in, best_out)
	      (IntSet.filter (fun v -> not (IntSet.do_intersect c_s (Graph.vertex_label g v))) n)
	      (-1, 0.0, infinity) in
	  if best_new = -1 then None
	  else
	    let s = IntSet.add s best_new in
	    let c_s = IntSet.union c_s (Graph.vertex_label g best_new) in
	    let n = IntSet.delete n best_new in
	    let n = IntSet.union n (IntSet.minus (Graph.neighbor_set g best_new) s) in
	    search s c_s n
      in
      if not (Graph.has_vertex g u) then loop g dk restore candidates else
      match search (IntSet.singleton u) (Graph.vertex_label g u) (Graph.neighbor_set g u) with
	| None -> loop g dk restore candidates
	| Some (s, n, cut) ->
	  (* Printf.eprintf "found colorful %a-bridge!\n%!" Util.output_float cut; *)
	  let g, dk, restore = IntSet.fold
	    (fun (g, dk, restore) v ->
	      let g, dk', restore' = Colcom.merge g u v in
	      g, dk +. dk', restore @@ restore')
	    (IntSet.delete s u) (g, dk, restore) in
	  let g = if Graph.deg g u = 0 then Graph.delete_vertex g u else g in
	  let candidates = IntSet.union candidates n in
	  (* Printf.eprintf "candidates = %a g = %a\n" IntSet.output candidates Graph.output' g; *)
	  loop g dk restore candidates
  in
  let g, dk, restore = loop g 0.0 (fun x -> x) candidates in
  {colcom with Colcom.g = g; Colcom.t_bridge_candidates = IntSet.empty}, dk, restore
;;

let forbidden_edge colcom =
  let g, dk, forbidden = 
    Graph.fold_edges
      (fun (g, dk, forbidden) u v w_uv ->
	if IntSet.do_intersect (Graph.vertex_label g u) (Graph.vertex_label g v)
	then
	  Graph.disconnect g u v,
	  dk +. w_uv,
	  (u, v) :: forbidden
	else g, dk, forbidden)
      colcom.Colcom.g (colcom.Colcom.g, 0.0, []) in
  {colcom with Colcom.g = g}, dk, fun del -> forbidden @ del
;;
