/* colcom -- solve the colorful components problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#ifndef CC_GRAPH_INCLUDED
#define CC_GRAPH_INCLUDED

#include <cassert>
#include <cstdint>
#include <climits>

#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <iostream>

class ColorSet {
public:
    typedef uint64_t colors_t;
    static const int MAX_COLORS = sizeof (colors_t) * CHAR_BIT;
    ColorSet() : colors_(0) { }
    ColorSet(uint64_t c) : colors_(c) { }

    class Iterator {
    public:
	Iterator(uint64_t c) : colors_(c) { }
	Iterator& operator++() { colors_ &= colors_ - 1; return *this; }
	int operator*() { return __builtin_ctzll(colors_); }
	bool operator!=(const Iterator other) { return colors_ != other.colors_; }
    private:
	colors_t colors_;
    };
    Iterator begin() const { return Iterator(colors_); }
    Iterator end() const { return Iterator(0); }

    bool empty() const { return colors_ == 0; }
    bool contains(int color) { return colors_ & (colors_t(1) << color); }
    int max() const { return MAX_COLORS - __builtin_clzll(colors_); }
    colors_t as_int() const { return colors_; }
    void add(int color) { assert(color < MAX_COLORS); colors_ |= colors_t(1) << color; }
    int pop() {
	int r = __builtin_ctzll(colors_);
	colors_ &= colors_ - 1;
	return r;
    }

    // safe "bool"
    operator const void*() const { return empty() ? 0 : this; }
	
    ColorSet operator&(ColorSet other) const { return ColorSet(colors_ & other.colors_); }
    bool operator!=(ColorSet other) const { return colors_ != other.colors_; }

    ColorSet operator|=(ColorSet other) { colors_ |= other.colors_; return *this; }

private:
    colors_t colors_;
};

class NeighborIt {
public:
    NeighborIt(std::unordered_map<int, double>::const_iterator it) : it_(it) { }
    bool operator!=(const NeighborIt other) { return it_ != other.it_; }
    NeighborIt& operator++() { ++it_; return *this; }
    int n() const { return it_->first; }
    double w() const { return it_->second; }
    
private:
    std::unordered_map<int, double>::const_iterator it_;
};

class Graph {
public:
    Graph(int n = 0) : colors_(n), neighbors_(n) { }

    int num_vertices() const { return neighbors_.size(); }
    int num_edges() const {
	int n = 0;
	for (std::size_t i = 0; i < neighbors_.size(); i++)
	    n += neighbors_[i].size();
	return n / 2;
    }
    int max_color() const {
	int m = 0;
	for (int u = 0; u < num_vertices(); u++)
	    m = std::max(m, colors(u).max());
	return m;
    }
    bool has_vertex(int u) const { return u >= 0 && u < num_vertices(); }
    ColorSet colors(int u) const { assert (has_vertex(u)); return colors_[u]; }
    bool is_connected(int u, int v) const {
	assert(has_vertex(u) && has_vertex(v));
	return neighbors_[u].find(v) != neighbors_[u].end();
    }
    double weight(int u, int v) const {
	assert(has_vertex(u) && has_vertex(v));
	return neighbors_[u].find(v)->second;
    }
    NeighborIt neighbors_begin(int u) const {
	assert(has_vertex(u));
	return neighbors_[u].begin();
    }
    NeighborIt neighbors_end(int u) const {
	assert(has_vertex(u));
	return neighbors_[u].end();
    }

    void add_vertex(ColorSet c) { colors_.push_back(c); neighbors_.push_back(std::unordered_map<int, double>()); }
    void setColor(int u, ColorSet c) { assert(has_vertex(u)); colors_[u] = c; }
    void connect(int u, int v, double w) {
	assert(has_vertex(u) && has_vertex(v));
	neighbors_[u][v] = w;
	neighbors_[v][u] = w;
    }
    void disconnect(int u, int v) {
	assert(is_connected(u, v));
	neighbors_[u].erase(neighbors_[u].find(v));
	neighbors_[v].erase(neighbors_[v].find(u));
    }
    void set_weight(int u, int v, double w) { connect(u, v, w); }

private:
    std::vector<ColorSet> colors_;
    std::vector<std::unordered_map<int, double> > neighbors_;    
};

inline std::ostream& operator<<(std::ostream& out, const Graph& g) {
    for (int u = 0; u < g.num_vertices(); u++) {
	out << "v " << u;
	ColorSet colors = g.colors(u);
	while (!colors.empty()) {
	    int c = colors.pop();
	    out << ' ' << c;
	}
	out << std::endl;
    }
    for (int u = 0; u < g.num_vertices(); u++) {
	for (NeighborIt it = g.neighbors_begin(u); it != g.neighbors_end(u); ++it)
	    if (it.n() >= u)
		out << "e " << u << ' ' << it.n()
		    << ' ' << it.w() << std::endl;
    }
    return out;
}

struct Edge {
    Edge() { }
    Edge(int nu, int nv) {
	if (nu < nv) {
	    u = nu;
	    v = nv;
	} else {
	    u = nv;
	    v = nu;
	}
    }
    bool operator==(const Edge& other) const { return u == other.u && v == other.v; }
    int u, v;
};

template <class T>
inline void hash_combine(std::size_t& seed, const T& v) {
    std::hash<T> hasher;
    seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

namespace std {
    template<>
    struct hash<Edge> {
	size_t operator()(const Edge& e) const {
	    size_t seed = 0;
	    hash_combine(seed, e.u);
	    hash_combine(seed, e.v);
	    return seed;
	}
    };
}

inline bool sim(double x, double y) {
    return abs(x - y) < 1e-8;
}

#if 1
inline std::ostream& operator<<(std::ostream& out, const Edge& e) {
    return out << '{' << e.u << ", " << e.v << '}';
}

template<typename T>
std::ostream& operator<<(std::ostream& out, const std::vector<T> v) {
    out << '{';
    for (std::size_t i = 0; i < v.size(); i++) {
	if (i)
	    out << ", ";
	out << '[' << i << ']' << v[i];
    }
    return out << '}';
}

template<typename T>
std::ostream& operator<<(std::ostream& out, const std::unordered_set<T> v) {
    out << '{';
    bool start = true;
    for (auto it = v.begin(); it != v.end(); ++it) {
	if (start)
	    start = false;
	else 
	    out << ", ";
	out << *it;
    }
    return out << '}';
}

template<typename K, typename V>
std::ostream& operator<<(std::ostream& out, const std::unordered_map<K, V> v) {
    out << '{';
    bool start = true;
    for (auto it = v.begin(); it != v.end(); ++it) {
	if (start)
	    start = false;
	else 
	    out << ", ";
	out << it->first << ':' << it->second;
    }
    return out << '}';
}

#endif

#endif
