/* colcom -- solve the colorful components problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#include <algorithm>
#include <iostream>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern "C" {
#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/memory.h>
#include <caml/mlvalues.h>
}

#include "cc_graph.hh"

size_t read_data(FILE* in, char **buf_out) {
    size_t buf_size = 8192;
    char *buffer = (char *) malloc(buf_size);
    size_t capacity = buf_size;
    size_t p = 0;
    size_t bytes_read;
    while ((bytes_read = fread(buffer + p, 1, capacity, in)) > 0) {
	p += bytes_read;
	capacity -= bytes_read;
	if (capacity == 0) {
	    buf_size *= 2;
	    buffer = (char *) realloc(buffer, buf_size);
	    capacity = buf_size - p;
	}
	//fprintf(stderr, "bytes_read = %zd, p = %zd, capacity = %zd, buf_size = %zd\n",
	//	bytes_read, p, capacity, buf_size);
		
    }
    buffer = (char *) realloc(buffer, p);
    *buf_out = buffer;
    return p;
}

struct str_eq {
  bool operator()(const char* s1, const char* s2) const { return strcmp(s1, s2) == 0; }
};

struct str_hash {
    size_t operator()(const char* s) const { 
	size_t h = 0;
	for (; *s; ++s)
	    h = 33 * h + *s;
	return h;
    }
};

struct Neighbor {
    Neighbor() { }
    Neighbor(int nn, double nw) : n(nn), w(nw) { }
    int n;
    double w;
};

void dfs(const Graph& g, std::unordered_set<int>& seen, std::unordered_set<int>& component, int i) {
    seen.insert(i);
    component.insert(i);
    for (NeighborIt it = g.neighbors_begin(i); it != g.neighbors_end(i); ++it) {
	if (seen.find(it.n()) == seen.end())
	    dfs(g, seen, component, it.n());
    }
}

extern "C" {

//int main() {
value read_and_fold_components(value v_f, value v_accu, value v_verbose) {
    CAMLparam3 (v_f, v_accu, v_verbose);
    CAMLlocal4 (v_vertices, v_edges, v_tuple, v_cell);

    bool verbose = Int_val(v_verbose);

    char *buffer;
    size_t n = read_data(stdin, &buffer);
    if (verbose)
	printf("read %zd bytes\n", n);

    typedef std::unordered_map<const char*, int, str_hash, str_eq> strmap;
    std::vector<ColorSet> colors;
    
    strmap vertex_numbers, color_numbers;
    std::vector<const char *> vertex_names;
    int next_color = 0;

    Graph g;
    
    char *p = buffer, *end = buffer + n;
    size_t lineno = 0, num_edges = 0;
    bool seen_an_edge = false;
    while (p != end) {
	++lineno;
	while (*p == ' ' || *p == '\t') p++;
	switch (*p++) {
	case '\n':
	    break;
	case '#':
	    while (*p != '\n') p++;
	    break;
	case 'v': {
	    p++;
	    while (*p == ' ' || *p == '\t') p++;
	    const char *name = p;
	    while (*p != ' ' && *p != '\t' && *p != '\n') p++;
	    bool done = *p == '\n';
	    *p++ = '\0';

	    ColorSet cs;
	    while (!done) {
		while (*p == ' ' || *p == '\t') p++;
		if (*p == '\n')
		    break;
		const char *color_name = p;
		while (*p != ' ' && *p != '\t' && *p != '\n') p++;
		if (*p == '\n')
		    done = true;
		*p++ = '\0';
		int color;
		std::pair<strmap::iterator, bool> ret
		    = color_numbers.insert(std::make_pair(color_name, next_color));
		if (ret.second) {
		    //std::cerr << "new color '" << color_name << "'\n";
		    color = next_color++;
		    if (color >= ColorSet::MAX_COLORS) {
			std::cerr << "error: too many colors\n";
			exit(1);
		    }
		} else {
		    color = ret.first->second;
		}
		cs.add(color);
	    }

	    std::pair<strmap::iterator, bool> ret = vertex_numbers.insert(std::make_pair(name, g.num_vertices()));
	    if (ret.second) { // new
		g.add_vertex(cs);
		vertex_names.push_back(name);
	    } else {
		ColorSet old_colors = g.colors(ret.first->second);
		if (old_colors != cs) {
		    std::cerr << "error: line " << lineno << ": vertex repeated with different colors\n";
		    exit(1);
		}
	    }
	    break;
	}
	case 'e': {
	    num_edges++;
	    if (!seen_an_edge) {
		if (verbose)
		    std::cerr << "parsed " << g.num_vertices() << " vertices with " << next_color << " colors\n";
		seen_an_edge = true;
	    }
	    //while (*p != '\n') p++;
	    while (*p == ' ' || *p == '\t') p++;
	    const char *v1 = p;
	    while (*p != ' ' && *p != '\t') {
		if (*p == '\n') {
		    std::cerr << "error: line " << lineno << ": syntax error1\n";
		    exit(1);
		}
		p++;
	    }
	    *p++ = '\0';
	    while (*p == ' ' || *p == '\t') p++;
	    const char *v2 = p;
	    if (*p == '\n') {
		std::cerr << "error: line " << lineno << ": syntax error2\n";
		exit(1);
	    }
	    while (*p != ' ' && *p != '\t' && *p != '\n')
		p++;

	    double weight = 1.0;
	    if (*p != '\n') {
		*p++ = '\0';
		while (*p == ' ' || *p == '\t') p++;
		if (*p != '\n') {
		    char *end;
		    errno = 0;
		    weight = strtod(p, &end);
		    if (errno || end == p || (*end != ' ' && *end != '\t' && *end != '\n')) {
			std::cerr << "error: line " << lineno << ": syntax error3\n";
			exit(1);
		    }
		    p = end;
		    while (*p == ' ' || *p == '\t') p++;
		    if (*p++ != '\n') {
			std::cerr << "error: line " << lineno << ": syntax error4\n";
			exit(1);
		    }
		}
	    } else {
		*p++ = '\0';
	    }
	    strmap::const_iterator iv1 = vertex_numbers.find(v1);
	    strmap::const_iterator iv2 = vertex_numbers.find(v2);
	    if (iv1 == vertex_numbers.end() || iv2 == vertex_numbers.end()) {
		std::cerr << "error: line " << lineno << ": unknown vertex\n";
		exit(1);
	    }
	    if (g.is_connected(iv1->second, iv2->second)) {
		if (g.weight(iv1->second, iv2->second) != weight) {
		    std::cerr << "error: line " << lineno << " already connected with different weight\n";
		    exit(1);
		}
	    } else {
		g.connect(iv1->second, iv2->second, weight);
	    }
	    break;
	}
	default:
	    std::cerr << "syntax error: line " << lineno << " starts with '" << *p << "'\n";
	    exit(1);
	}
    }

    if (verbose)
	std::cerr << "parsed " << num_edges << " edges\n";

    std::unordered_set<int> seen;
    int num_components = 0;
    //int num_noncolorful_components = 0;
    std::vector<std::unordered_set<int> > components;//, colorful_components;
    for (int i = 0; i < g.num_vertices(); i++) {
	if (seen.find(i) == seen.end()) {
	    std::unordered_set<int> component;
	    dfs(g, seen, component, i);
	    num_components++;

#if 1
	    components.push_back(component);
#else
	    bool is_colorful = true;
	    ColorSet cs;
	    for (std::unordered_set<int>::const_iterator it = component.begin(); it != component.end(); ++it) {
		if (cs & g.colors(*it)) {
		    num_noncolorful_components++;
		    is_colorful = false;
		    break;
		}
		cs |= g.colors(*it);
	    }
	    if (is_colorful) {
		if (print_components)
		    colorful_components.push_back(component);
	    } else {
		components.push_back(component);
	    }
#endif
	}
    }

    struct {
	bool operator()(const std::unordered_set<int>& c1,
			const std::unordered_set<int>& c2) const {
	    return c1.size() > c2.size();
	}
    } cmp;

    if (verbose)
	std::cout << "found " << num_components << " connected components\n";
    std::sort(components.begin(), components.end(), cmp);

    for (size_t i = 0; i < components.size(); i++) {
	std::unordered_map<int, int> new_numbers, old_numbers;

	v_vertices = caml_alloc(components[i].size(), 0);
	int u = 0;
	for (std::unordered_set<int>::const_iterator it = components[i].begin(); it != components[i].end(); ++it) {
	    new_numbers[*it] = u;
	    old_numbers[u] = *it;

	    v_tuple = caml_alloc(2, 0);
	    Store_field(v_tuple, 0, Val_int(g.colors(*it).as_int()));
	    Store_field(v_tuple, 1, caml_copy_string(vertex_names[*it]));
	    Store_field(v_vertices, u, v_tuple);
	    u++;
	}

	v_edges = Val_int(0);	// empty list
	for (std::unordered_set<int>::const_iterator it = components[i].begin(); it != components[i].end(); ++it) {
	    for (NeighborIt nit = g.neighbors_begin(*it); nit != g.neighbors_end(*it); ++nit) {
		int u = new_numbers[*it];
		int v = new_numbers[nit.n()];
		if (!(u < v))
		    continue;
		double w = nit.w();
		v_tuple = caml_alloc(3, 0);
		Store_field(v_tuple, 0, Val_int(u));
		Store_field(v_tuple, 1, Val_int(v));
		Store_field(v_tuple, 2, caml_copy_double(w));
		v_cell = caml_alloc(2, 0);
		Store_field(v_cell, 0, v_tuple);
		Store_field(v_cell, 1, v_edges);
		v_edges = v_cell;
	    }
	}

	v_accu = caml_callback3(v_f, v_accu, v_vertices, v_edges);
    }

    CAMLreturn(v_accu);
}

}
